import asyncio
from typing import Type

from asgiref.typing import ASGI3Application

from .global_state import GlobalState
from .iglobal_state import IGlobalState
from .iprotocol import IProtocol
from .isocket_provider import ISocketProvider


class Server:
    socket_provider: ISocketProvider
    protocol_class: Type[IProtocol]
    server: asyncio.Server | None
    app: ASGI3Application
    global_state: IGlobalState

    def __init__(
        self,
        app: ASGI3Application,
        socket_provider: ISocketProvider,
        protocol_class: Type[IProtocol],
        global_state: IGlobalState | None = None,
    ) -> None:
        if global_state is None:
            global_state = GlobalState()
        self.socket_provider = socket_provider
        self.protocol_class = protocol_class
        self.global_state = global_state
        self.server = None
        self.app = app

    async def init(self) -> None:
        loop = asyncio.get_running_loop()
        sock = self.socket_provider.acquire()
        self.server = await loop.create_server(
            lambda: self.protocol_class(self.global_state, self.app),
            sock=sock,
            start_serving=False,
        )

    async def close(self) -> None:
        if self.server is not None and self.server.is_serving():
            self.server.close()
            await self.server.wait_closed()
        self.socket_provider.cleanup()

    async def start_serving(self) -> None:
        if self.server is None:
            raise ValueError("Server is not initialized yet")
        await self.server.start_serving()



import socket
from abc import ABC, abstractmethod
from typing import Literal


INET_FAMILY = (
    Literal[socket.AddressFamily.AF_INET]
    | Literal[socket.AddressFamily.AF_INET6]
    | Literal[socket.AddressFamily.AF_UNSPEC]
)


class ISocketProvider(ABC):
    @abstractmethod
    def acquire(self) -> socket.socket:
        raise NotImplementedError

    @abstractmethod
    def cleanup(self) -> None:
        raise NotImplementedError



import os
import socket

from .isocket_provider import INET_FAMILY, ISocketProvider


class InetSocketProvider(ISocketProvider):
    host: str
    port: int
    family: INET_FAMILY
    sock: socket.socket | None

    def __init__(
        self,
        host: str,
        port: int,
        family: INET_FAMILY = socket.AddressFamily.AF_UNSPEC,
    ) -> None:
        self.host = host
        self.port = port
        self.family = family
        self.sock = None

    def acquire(self) -> socket.socket:
        if self.sock is None:
            self.sock = self.create_socket()
        return self.sock

    def create_socket(self) -> socket.socket:
        sock = socket.socket(self.family, socket.SOCK_STREAM)
        self.configure_socket(sock)
        sock.bind((self.host, self.port))
        return sock

    def configure_socket(self, sock: socket.socket) -> None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def cleanup(self) -> None:
        pass


class UnixSocketProvider(ISocketProvider):
    path: str
    sock: socket.socket | None

    def __init__(self, path: str) -> None:
        self.path = path
        self.sock = None

    def acquire(self) -> socket.socket:
        if self.sock is not None:
            return self.sock
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.configure_socket(self.sock)
        self.sock.bind(self.path)
        return self.sock

    def configure_socket(self, sock: socket.socket) -> None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def cleanup(self) -> None:
        os.unlink(self.path)



import asyncio
from dataclasses import dataclass
from typing import Iterable, Literal

import httptools

from .utils import get_remote_addr, is_ssl


@dataclass
class HTTPRequest:
    path: str
    method: str
    root_path: str
    raw_path: bytes
    http_version: str
    query_string: bytes
    client: tuple[str, int] | None
    server: tuple[str, int] | None
    headers: Iterable[tuple[bytes, bytes]]
    scheme: Literal["https"] | Literal["http"]


class HTTPRequestParser:
    request: HTTPRequest | None
    url: bytes | None
    body: bytes | None
    body_event: asyncio.Event
    request_event: asyncio.Event
    headers: list[tuple[bytes, bytes]]
    is_completed: bool
    transport: asyncio.Transport

    def __init__(
        self,
        transport: asyncio.Transport,
    ) -> None:
        self.parser = httptools.HttpRequestParser(self)
        self.transport = transport
        self.request = None
        self.url = None
        self.body = None
        self.headers = []
        self.body_event = asyncio.Event()
        self.request_event = asyncio.Event()
        self.more_body = True

    def disconnect(self) -> None:
        self.more_body = False

    def on_url(self, url: bytes) -> None:
        self.url = url

    def on_header(self, name: bytes, value: bytes) -> None:
        self.headers.append((name.decode().lower().encode(), value))

    def on_headers_complete(self) -> None:
        if self.parser.should_upgrade():
            self.transport.pause_reading()
            self.transport.write(
                b"HTTP/1.1 501\n\n<h1>Server does"
                b" not support websockets connections</h1>"
            )
            self.transport.write_eof()
            return
        assert self.url is not None
        url_object = httptools.parse_url(self.url)
        self.request = HTTPRequest(
            http_version=self.parser.get_http_version().upper(),
            scheme="http" if not is_ssl(self.transport) else "https",
            path=url_object.path.decode(),
            raw_path=self.url,
            query_string=url_object.query,
            root_path="",
            headers=self.headers,
            server=("localhost", 8000),
            client=get_remote_addr(self.transport),
            method=self.parser.get_method().decode(),
        )
        self.request_event.set()
        self.transport.pause_reading()

    async def get_request(self) -> HTTPRequest:
        if self.request is None:
            await self.request_event.wait()
            self.request_event.clear()
            assert self.request is not None
        return self.request

    def on_body(self, body: bytes) -> None:
        if self.body is None:
            self.body = b""
        self.body += body
        self.body_event.set()

    async def receive_body(self) -> tuple[bytes | None, bool]:
        if self.body is None:
            self.body = b""
            return (self.body, False)
        if self.more_body is True or self.body == b"":
            self.transport.resume_reading()
            await self.body_event.wait()
            self.body_event.clear()
        body = self.body
        self.body = b""
        return (body, self.more_body)

    def feed_data(self, data: bytes) -> None:
        self.parser.feed_data(data)

    def on_message_complete(self) -> None:
        self.more_body = False



import asyncio

from asgiref.typing import ASGI3Application

from .asgi_controller import ASGIController
from .iglobal_state import IGlobalState
from .iprotocol import IProtocol
from .request_parser import HTTPRequestParser
from .response_serializer import HTTPResponseSerializer


class HTTPProtocol(IProtocol):
    def __init__(
        self, global_state: IGlobalState, app: ASGI3Application
    ) -> None:
        self.app = app
        self.global_state = global_state

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        if not isinstance(transport, asyncio.Transport):
            raise ValueError("transport must be instance of asyncio.Transport")
        self.request_parser = HTTPRequestParser(transport)
        self.asgi_controller = ASGIController(
            app=self.app,
            request_parser=self.request_parser,
            response_serializer=HTTPResponseSerializer(transport),
        )
        self.global_state.add_task(
            asyncio.create_task(self.asgi_controller.start())
        )

    def connection_lost(self, exc: Exception | None) -> None:
        self.request_parser.disconnect()
        if exc is not None:
            print(exc)

    def data_received(self, data: bytes) -> None:
        self.request_parser.feed_data(data)

    def eof_received(self) -> bool:
        return False



import asyncio


def is_ssl(transport: asyncio.Transport) -> bool:
    return transport.get_extra_info("sslcontext") is None


def get_remote_addr(transport: asyncio.Transport) -> tuple[str, int] | None:
    if socket_info := transport.get_extra_info("socket"):
        if info := socket_info.getpeername():
            return (str(info[0]), int(info[1]))
    if info := transport.get_extra_info("peername"):
        return (str(info[0]), int(info[1]))
    return None



import asyncio
import enum
from typing import Iterable, cast

from asgiref.typing import (
    ASGISendEvent,
    HTTPResponseBodyEvent,
    HTTPResponseStartEvent,
    HTTPResponseTrailersEvent,
)


class HTTPResponseEvents(enum.Enum):
    START = "http.response.start"
    TRAILERS = "http.response.trailers"
    BODY = "http.response.body"


class HTTPResponseSerializer:
    expected_event: HTTPResponseEvents | None
    transport: asyncio.WriteTransport

    def __init__(self, transport: asyncio.WriteTransport) -> None:
        self.expected_event = HTTPResponseEvents.START
        self.transport = transport

    def is_completed(self) -> bool:
        return self.expected_event is None

    def send(self, event: ASGISendEvent) -> None:
        self.validate_event(event)
        match event["type"]:
            case HTTPResponseEvents.START.value:
                event = cast(HTTPResponseStartEvent, event)
                headers = self.encode_headers(event["headers"])
                message: bytes = (
                    f'HTTP/1.1 {event["status"]}\n'.encode() + headers
                )
                if event.get("trailers", False) is True:
                    self.expected_event = HTTPResponseEvents.TRAILERS
                else:
                    self.expected_event = HTTPResponseEvents.BODY
                    message += b"\n"
                self.transport.write(message)
            case HTTPResponseEvents.TRAILERS.value:
                event = cast(HTTPResponseTrailersEvent, event)
                message = self.encode_headers(event["headers"])
                if event.get("more_trailers", False) is False:
                    message += b"\n"
                    self.expected_event = HTTPResponseEvents.BODY
                self.transport.write(message)
            case HTTPResponseEvents.BODY.value:
                event = cast(HTTPResponseBodyEvent, event)
                message = event["body"]
                self.transport.write(message)
                if event.get("more_body", False) is False:
                    self.expected_event = None
                    self.transport.write_eof()
            case _:
                self.transport.write_eof()
                raise RuntimeError(f"Unhandled event type: {event['type']}")

    def encode_headers(self, headers: Iterable[tuple[bytes, bytes]]) -> bytes:
        return b"".join(map(lambda h: h[0] + b": " + h[1] + b"\n", headers))

    def validate_event(self, event: ASGISendEvent) -> None:
        assert self.expected_event is not None, "Response already ended"
        assert event["type"] == self.expected_event.value



import asyncio
from typing import Any

from .iglobal_state import IGlobalState


class GlobalState(IGlobalState):
    tasks: set[asyncio.Task[Any]]

    def __init__(self) -> None:
        self.tasks = set()

    def add_task(self, task: asyncio.Task[Any]) -> None:
        self.tasks.add(task)
        task.add_done_callback(self.remove_task)

    def remove_task(self, task: asyncio.Task[Any]) -> None:
        self.tasks.remove(task)



import asyncio
from abc import ABC, abstractmethod
from typing import Any


class IGlobalState(ABC):
    @abstractmethod
    def add_task(self, task: asyncio.Task[Any]) -> None:
        raise NotImplementedError



from .global_state import GlobalState
from .http_protocol import HTTPProtocol
from .iglobal_state import IGlobalState
from .iprotocol import IProtocol
from .isocket_provider import ISocketProvider
from .server import Server as Favicorn
from .socket_providers import InetSocketProvider, UnixSocketProvider

__all__ = (
    "Favicorn",
    "HTTPProtocol",
    "ISocketProvider",
    "InetSocketProvider",
    "UnixSocketProvider",
    "IProtocol",
    "IGlobalState",
    "GlobalState",
)



import asyncio
from abc import abstractmethod

from asgiref.typing import ASGI3Application

from .iglobal_state import IGlobalState


class IProtocol(asyncio.Protocol):
    @abstractmethod
    def __init__(
        self, global_state: IGlobalState, app: ASGI3Application
    ) -> None:
        raise NotImplementedError



from asgiref.typing import (
    ASGI3Application,
    ASGIReceiveEvent,
    ASGISendEvent,
    ASGIVersions,
    HTTPDisconnectEvent,
    HTTPRequestEvent,
    HTTPResponseBodyEvent,
    HTTPResponseStartEvent,
    HTTPScope,
)

from .request_parser import HTTPRequestParser
from .response_serializer import HTTPResponseSerializer


class ASGIController:
    app: ASGI3Application
    body: bytes | None
    request_parser: HTTPRequestParser
    response_serializer: HTTPResponseSerializer

    def __init__(
        self,
        app: ASGI3Application,
        request_parser: HTTPRequestParser,
        response_serializer: HTTPResponseSerializer,
    ) -> None:
        self.app = app
        self.request_parser = request_parser
        self.response_serializer = response_serializer

    async def start(self) -> None:
        request = await self.request_parser.get_request()
        scope = HTTPScope(
            type="http",
            asgi=ASGIVersions(spec_version="3.0", version="3.0"),
            http_version=request.http_version,
            scheme=request.scheme,
            path=request.path,
            raw_path=request.raw_path,
            query_string=request.query_string,
            root_path=request.root_path,
            headers=request.headers,
            server=request.server,
            client=request.client,
            extensions={},
            method=request.method,
        )
        try:
            await self.app(scope, self.receive, self.send)
        except BaseException as unhandled_error:
            print(unhandled_error)
            await self.send_500_response()

    async def receive(self) -> ASGIReceiveEvent:
        body, more_body = await self.request_parser.receive_body()
        if body is None:
            return HTTPDisconnectEvent(type="http.disconnect")
        return HTTPRequestEvent(
            type="http.request",
            body=body,
            more_body=more_body,
        )

    async def send(self, event: ASGISendEvent) -> None:
        self.response_serializer.send(event)

    async def send_500_response(self) -> None:
        await self.send(
            HTTPResponseStartEvent(
                type="http.response.start",
                status=500,
                headers=[],
                trailers=False,
            )
        )
        await self.send(
            HTTPResponseBodyEvent(
                type="http.response.body",
                body=b"Internal Server Error",
                more_body=False,
            )
        )
